@@ -8,12 +8,12 @@ use warnings;
 use App::Music::ChordPro::Chords;
 use App::Music::ChordPro::Output::Common;
 
-use Encode qw(decode encode);
+use Encode qw(decode decode_utf8 encode);
 use Carp;
 
 sub new {
     my ($pkg) = @_;
-    bless { songs => [ App::Music::ChordPro::Song->new ] }, $pkg;
+    bless { songs => [ ] }, $pkg;
 }
 
 # Parser context.
@@ -62,7 +62,9 @@ sub parsefile {
 	$data = do { local $/; <STDIN> };
     }
     else {
-	open( my $fh, '<', $filename)
+	my $name = $filename;
+	$filename = decode_utf8($name);
+	open( my $fh, '<', $name)
 	  or croak("$filename: $!\n");
 	$data = do { local $/; <$fh> };
     }
@@ -113,24 +115,50 @@ sub parsefile {
 	}
     }
 
+    $diag->{format} = $options->{diagformat}
+      || $::config->{diagnostics}->{format};
+    $diag->{file} = $filename;
+
     # Split in lines;
-    my @lines = split( /\r\n|\n|\r/, $data );
+    my @lines;
+    $data =~ s/^\s+//s;
+    # Unless empty, make sure there is a final newline.
+    $data .= "\n" if $data =~ /.(?!\r\n|\n|\r)\Z/;
+    # We need to maintain trailing newlines.
+    push( @lines, $1 ) while $data =~ /(.*)(?:\r\n|\n|\r)/g;
+
+    my $linecnt = 0;
+    while ( @lines ) {
+	my $song = $self->parse_song( \@lines, \$linecnt, $options );
+#	if ( exists($self->{songs}->[-1]->{body}) ) {
+	    push( @{ $self->{songs} }, $song );
+#	}
+#	else {
+#	    $self->{songs} = [ $song ];
+#	}
+    }
+    return 1;
+}
+
+my $song;			# current song
+
+sub parse_song {
+    my ( $self, $lines, $linecnt, $options ) = @_;
 
     $no_transpose = $options->{'no-transpose'};
     $no_substitute = $options->{'no-substitute'};
 
-    push( @{ $self->{songs} }, App::Music::ChordPro::Song->new )
-      if exists($self->{songs}->[-1]->{body});
-    $self->{songs}->[-1]->{structure} = "linear";
+    $song = App::Music::ChordPro::Song->new
+      ( source => { file => $diag->{file}, line => 1 + $$linecnt },
+	structure => "linear",
+      );
+
     $xpose = 0;
     $grid_arg = '1+4x4+1';
     $in_context = $def_context;
     @used_chords = ();
     %warned_chords = ();
     App::Music::ChordPro::Chords::reset_song_chords();
-    $diag->{format} = $options->{diagformat}
-      || $::config->{diagnostics}->{format};
-    $diag->{file} = $filename;
 
     # Build regex for the known metadata items.
     if ( $::config->{metadata}->{keys} ) {
@@ -143,18 +171,22 @@ sub parsefile {
 	undef $re_meta;
     }
 
-    my $linecnt;
-    while ( @lines ) {
-	$diag->{line} = ++$linecnt;
-	$diag->{orig} = $_ = shift(@lines);
+    while ( @$lines ) {
+	$diag->{line} = ++$$linecnt;
+	$diag->{orig} = $_ = shift(@$lines);
+
+	if ( /^\s*\{(new_song|ns)\}\s*$/ ) {
+	    last if $song->{body};
+	    next;
+	}
 
 	if ( /^#/ ) {
 	    # Collect pre-title stuff separately.
-	    if ( exists $self->{songs}->[-1]->{title} ) {
+	    if ( exists $song->{title} ) {
 		$self->add( type => "ignore", text => $_ );
 	    }
 	    else {
-		push( @{ $self->{songs}->[-1]->{preamble} }, $_ );
+		push( @{ $song->{preamble} }, $_ );
 	    }
 	    next;
 	}
@@ -183,20 +215,20 @@ sub parsefile {
 	if ( /\S/ ) {
 	    $self->add( type => "songline", $self->decompose($_) );
 	}
-	elsif ( exists $self->{songs}->[-1]->{title} ) {
+	elsif ( exists $song->{title} ) {
 	    $self->add( type => "empty" );
 	}
 	else {
 	    # Collect pre-title stuff separately.
-	    push( @{ $self->{songs}->[-1]->{preamble} }, $_ );
+	    push( @{ $song->{preamble} }, $_ );
 	}
     }
     do_warn("Unterminated context in song: $in_context")
       if $in_context;
 
     my $diagrams;
-    if ( exists($self->{songs}->[-1]->{settings}->{diagrams} ) ) {
-	$diagrams = $self->{songs}->[-1]->{settings}->{diagrams};
+    if ( exists($song->{settings}->{diagrams} ) ) {
+	$diagrams = $song->{settings}->{diagrams};
 	$diagrams &&= $::config->{diagrams}->{show} || "all";
     }
     else {
@@ -224,24 +256,25 @@ sub parsefile {
 	    @used_chords =
 	      sort App::Music::ChordPro::Chords::chordcompare @used_chords;
 	}
-
-	$self->add( type   => "diagrams",
-		    origin => "song",
-		    show   => $diagrams,
-		    chords => [ @used_chords ] );
+	$song->{chords} =
+	  { type   => "diagrams",
+	    origin => "song",
+	    show   => $diagrams,
+	    chords => [ @used_chords ],
+	  };
     }
 
     # Global transposition.
-    $self->{songs}->[-1]->transpose( $options->{transpose} );
+    $song->transpose( $options->{transpose} );
 
-    # $self->{songs}->[-1]->structurize;
+    # $song->structurize;
 
-    return 1;
+    return $song;
 }
 
 sub add {
     my $self = shift;
-    push( @{$self->{songs}->[-1]->{body}},
+    push( @{$song->{body}},
 	  { context => $in_context,
 	    @_ } );
     push( @chorus, { context => $in_context, @_ } )
@@ -328,7 +361,7 @@ sub decompose {
 
 sub cdecompose {
     my ( $self, $line ) = @_;
-    $line = App::Music::ChordPro::Output::Common::fmt_subst( $self->{songs}->[-1],
+    $line = App::Music::ChordPro::Output::Common::fmt_subst( $song,
 						     $line )
       unless $no_substitute;
     my %res = $self->decompose($line);
@@ -448,6 +481,11 @@ sub directive {
 	    $grid_arg = $arg;
 	    $grid_cells = [ $2 * ( $3//1 ), ($1//0), ($4//0) ];
 	}
+	elsif ( $arg && $arg ne "" ) {
+	    $self->add( type  => "control",
+			name  => "tag",
+			value => $arg );
+	}
 	else {
 	    do_warn("Garbage in start_of_$1: $arg (ignored)\n")
 	      if $arg;
@@ -477,8 +515,6 @@ sub directive {
 
     # Song settings.
 
-    my $cur = $self->{songs}->[-1];
-
     # Breaks.
 
     if ( $dir =~ /^(?:colb|column_break)$/i ) {
@@ -492,9 +528,7 @@ sub directive {
     }
 
     if ( $dir =~ /^(?:new_song|ns)$/i ) {
-	return unless $self->{songs}->[-1]->{body};
-	push(@{$self->{songs}}, App::Music::ChordPro::Song->new );
-	return;
+	die("FATAL - cannot start a new song now\n");
     }
 
     # Comments. Strictly speaking they do not belong here.
@@ -558,14 +592,14 @@ sub directive {
     }
 
     if ( $dir =~ /^(?:title|t)$/ ) {
-	$cur->{title} = $arg;
-	push( @{ $self->{songs}->[-1]->{meta}->{title} }, $arg );
+	$song->{title} = $arg;
+	push( @{ $song->{meta}->{title} }, $arg );
 	return;
     }
 
     if ( $dir =~ /^(?:subtitle|st)$/ ) {
-	push(@{$cur->{subtitle}}, $arg);
-	push( @{ $self->{songs}->[-1]->{meta}->{subtitle} }, $arg );
+	push( @{ $song->{subtitle} }, $arg );
+	push( @{ $song->{meta}->{subtitle} }, $arg );
 	return;
     }
 
@@ -575,7 +609,7 @@ sub directive {
 	if ( $xpose && $1 eq "key" ) {
 	    $arg = App::Music::ChordPro::Chords::transpose( $arg, $xpose );
 	}
-	push( @{ $self->{songs}->[-1]->{meta}->{$1} }, $arg );
+	push( @{ $song->{meta}->{$1} }, $arg );
 	return;
     }
 
@@ -589,7 +623,7 @@ sub directive {
 	    }
 	    if ( $re_meta && $key =~ $re_meta ) {
 		# Known.
-		push( @{ $self->{songs}->[-1]->{meta}->{$key} }, $val );
+		push( @{ $song->{meta}->{$key} }, $val );
 	    }
 	    elsif ( $::config->{metadata}->{strict} ) {
 		# Unknown, and strict.
@@ -597,7 +631,7 @@ sub directive {
 	    }
 	    else {
 		# Allow.
-		push( @{ $self->{songs}->[-1]->{meta}->{$key} }, $val );
+		push( @{ $song->{meta}->{$key} }, $val );
 	    }
 	}
 	else {
@@ -619,34 +653,32 @@ sub global_directive {
     my ($self, $d, $legacy ) = @_;
     my ( $dir, $arg ) = dir_split($d);
 
-    my $cur = $self->{songs}->[-1];
-
     # Song / Global settings.
 
     if ( $dir eq "titles"
 	 && $arg =~ /^(left|right|center|centre)$/i ) {
-	$cur->{settings}->{titles} =
+	$song->{settings}->{titles} =
 	  lc($1) eq "centre" ? "center" : lc($1);
 	return 1;
     }
 
     if ( $dir eq "columns"
 	 && $arg =~ /^(\d+)$/ ) {
-	$cur->{settings}->{columns} = $arg;
+	$song->{settings}->{columns} = $arg;
 	return 1;
     }
 
     if ( $dir eq "pagetype" || $dir eq "pagesize" ) {
-	$cur->{settings}->{papersize} = $arg;
+	$song->{settings}->{papersize} = $arg;
 	return 1;
     }
 
     if ( $dir =~ /^(?:grid|g)$/ ) {
-	$cur->{settings}->{diagrams} = 1;
+	$song->{settings}->{diagrams} = 1;
 	return 1;
     }
     if ( $dir =~ /^(?:no_grid|ng)$/ ) {
-	$cur->{settings}->{diagrams} = 0;
+	$song->{settings}->{diagrams} = 0;
 	return 1;
     }
 
@@ -658,7 +690,7 @@ sub global_directive {
 		  name => "transpose",
 		  previous => $xpose,
 		);
-	my $m = $self->{songs}->[-1]->{meta};
+	my $m = $song->{meta};
 	if ( $m->{key} ) {
 	    $m->{key_actual} =
 	      [ App::Music::ChordPro::Chords::transpose( $m->{key}->[-1],
@@ -678,7 +710,7 @@ sub global_directive {
 		  name => "transpose",
 		  previous => $xpose,
 		);
-	my $m = $self->{songs}->[-1]->{meta};
+	my $m = $song->{meta};
 	if ( $m->{key} ) {
 	    $m->{key_from} =
 	      [ App::Music::ChordPro::Chords::transpose( $m->{key}->[-1],
@@ -871,8 +903,8 @@ sub global_directive {
 		$ci = $res->{name};
 	    }
 	    # Combine consecutive entries.
-	    if ( $self->{songs}->[-1]->{body}->[-1]->{type} eq "diagrams" ) {
-		push( @{ $self->{songs}->[-1]->{body}->[-1]->{chords} },
+	    if ( $song->{body}->[-1]->{type} eq "diagrams" ) {
+		push( @{ $song->{body}->[-1]->{chords} },
 		      $ci );
 	    }
 	    else {
@@ -884,7 +916,7 @@ sub global_directive {
 	}
 	elsif ( $res->{frets} || $res->{base} || $res->{fingers} ) {
 	    $res->{base} ||= 1;
-	    push( @{$cur->{define}}, $res );
+	    push( @{$song->{define}}, $res );
 	    if ( $res->{frets} ) {
 		my $res =
 		  App::Music::ChordPro::Chords::add_song_chord
